'use strict';

var TextDecoder = require("text-encoding").TextDecoder;

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var path = _interopDefault(require('path-source'));
var array = _interopDefault(require('array-source'));
var stream = _interopDefault(require('stream-source'));
var slice = _interopDefault(require('slice-source'));

function dbf_cancel() {
  return this._source.cancel();
}

function readBoolean(value) {
  return /^[nf]$/i.test(value) ? false
      : /^[yt]$/i.test(value) ? true
      : null;
}

function readDate(value) {
  return new Date(+value.substring(0, 4), value.substring(4, 6) - 1, +value.substring(6, 8));
}

function readNumber(value) {
  return isNaN(value = +value) ? null : value;
}

function readString(value) {
  return value.trim() || null;
}

var types = {
  B: readNumber,
  C: readString,
  D: readDate,
  F: readNumber,
  L: readBoolean,
  M: readNumber,
  N: readNumber
};

function dbf_read() {
  var that = this, i = 1;
  return that._source.slice(that._recordLength).then(function(value) {
    return value ? {done: false, value: that.fields.map(function(f) {
      return types[f.type](that._decode(value.subarray(i, i += f.length)));
    })} : {done: true, value: undefined};
  });
}

function view(array) {
  return new DataView(array.buffer, array.byteOffset, array.byteLength);
}

function dbf(source, decoder) {
  source = slice(source);
  return source.slice(32).then(function(array) {
    var head = view(array);
    return source.slice(head.getUint16(8, true) - 32).then(function(array) {
      return new Dbf(source, decoder, head, view(array));
    });
  });
}

function Dbf(source, decoder, head, body) {
  this._source = source;
  this._decode = decoder.decode.bind(decoder);
  this._recordLength = head.getUint16(10, true);
  this.version = head.getUint8(0);
  this.date = new Date(1900 + head.getUint8(1), head.getUint8(2) - 1, head.getUint8(3));
  this.length = head.getUint32(4, true);
  this.fields = [];
  for (var n = 0; body.getUint8(n) !== 0x0d; n += 32) {
    for (var j = 0; j < 11; ++j) if (body.getUint8(n + j) === 0) break;
    this.fields.push({
      name: this._decode(new Uint8Array(body.buffer, body.byteOffset + n, j)),
      type: String.fromCharCode(body.getUint8(n + 11)),
      length: body.getUint8(n + 16)
    });
  }
}

var prototype$1 = Dbf.prototype;
prototype$1.read = dbf_read;
prototype$1.cancel = dbf_cancel;

function cancel() {
  return this._source.cancel();
}

function readNull() {
  return {shapeType: 0};
};

function shp_read() {
  var that = this;
  return that._source.slice(8).then(function(array) {
    if (array == null) return {done: true, value: undefined};
    var header = view(array);
    return that._source.slice(header.getInt32(4, false) * 2).then(function(array) {
      var record = view(array);
      return {done: false, value: record.getInt32(0, true) ? that._type(record) : readNull()};
    });
  });
}

function readPoint(record) {
  return {shapeType: 1, x: record.getFloat64(4, true), y: record.getFloat64(12, true)};
};

function readPoly(shapeType) {
  return function(record) {
    var i = 44, j, n = record.getInt32(36, true), m = record.getInt32(40, true), parts = new Array(n), points = new Array(m);
    for (j = 0; j < n; ++j, i += 4) parts[j] = record.getInt32(i, true);
    for (j = 0; j < m; ++j, i += 16) points[j] = [record.getFloat64(i, true), record.getFloat64(i + 8, true)];
    return {shapeType: shapeType, box: [record.getFloat64(4, true), record.getFloat64(12, true), record.getFloat64(20, true), record.getFloat64(28, true)], parts: parts, points: points};
  };
};

function readMultiPoint(record) {
  var i = 40, j, n = record.getInt32(36, true), points = new Array(n);
  for (j = 0; j < n; ++j, i += 16) points[j] = [record.getFloat64(i, true), record.getFloat64(i + 8, true)];
  return {shapeType: 8, box: [record.getFloat64(4, true), record.getFloat64(12, true), record.getFloat64(20, true), record.getFloat64(28, true)], points: points};
};

var readPolyLine = readPoly(3);
var readPolygon = readPoly(5);
var types$1 = {
  0: readNull,
  1: readPoint,
  3: readPolyLine,
  5: readPolygon,
  8: readMultiPoint,
  11: readPoint,
  13: readPolyLine,
  15: readPolygon,
  18: readMultiPoint
  // TODO 21 readPointM
  // TODO 23 readPolyLineM
  // TODO 25 readPolygonM
  // TODO 28 readMultiPointM
  // TODO 31 readMultiPatch
};

function shp(source) {
  source = slice(source);
  return source.slice(100).then(function(array) {
    return new Shp(source, view(array));
  });
};

function Shp(source, header) {
  this.shapeType = header.getInt32(32, true);
  if (!(this.shapeType in types$1)) throw new Error("unsupported shape type: " + this.shapeType);
  this._source = source;
  this._type = types$1[this.shapeType];
  this.fileCode = header.getInt32(0, false);
  this.version =  header.getInt32(28, true);
  this.box = [header.getFloat64(36, true), header.getFloat64(44, true), header.getFloat64(52, true), header.getFloat64(60, true)];
}

var prototype$2 = Shp.prototype;
prototype$2.read = shp_read;
prototype$2.cancel = cancel;

function properties(fields) {
  return new Function("d", "return {"
      + fields.map(function(field, i) { return JSON.stringify(field.name) + ":d[" + i + "]"; })
      + "};");
}

function noproperties() {
  return {};
}

var shapeTypes = {
  0: none,
  1: point,
  3: polyline,
  5: polygon,
  8: multipoint,
  11: point, // PointZ
  13: polyline, // PolyLineZ
  15: polygon, // PolygonZ
  18: multipoint // MultiPointZ
};

function none() {
  return null;
}

function geometry(shapeType) {
  if (!(shapeType in shapeTypes)) throw new Error("unknown shape type: " + shapeType);
  var convert = shapeTypes[shapeType];
  return function(record) { return record.shapeType ? convert(record) : null; };
}

function point(record) {
  return {type: "Point", coordinates: [record.x, record.y]};
}

function multipoint(record) {
  return {type: "MultiPoint", coordinates: record.points};
}

function polyline(record) {
  return record.parts.length === 1
      ? {type: "LineString", coordinates: record.points}
      : {type: "MultiLineString", coordinates: record.parts.map(function(i, j) { return record.points.slice(i, record.parts[j + 1]); })};
}

function polygon(record) {
  var parts = record.parts.map(function(i, j) { return record.points.slice(i, record.parts[j + 1]); }),
      polygons = [],
      holes = [];

  parts.forEach(function(part) {
    if (ringClockwise(part)) polygons.push([part]);
    else holes.push(part);
  });

  holes.forEach(function(hole) {
    var point = hole[0];
    polygons.some(function(polygon) {
      if (ringContains(polygon[0], point)) {
        polygon.push(hole);
        return true;
      }
    }) || polygons.push([hole]);
  });

  return polygons.length > 1
      ? {type: "MultiPolygon", coordinates: polygons}
      : {type: "Polygon", coordinates: polygons[0]};
}

function ringClockwise(ring) {
  if ((n = ring.length) < 4) return false;
  var i = 0, n, area = ring[n - 1][1] * ring[0][0] - ring[n - 1][0] * ring[0][1];
  while (++i < n) area += ring[i - 1][1] * ring[i][0] - ring[i - 1][0] * ring[i][1];
  return area >= 0;
}

function ringContains(ring, point) {
  var x = point[0], y = point[1], contains = false;
  for (var i = 0, n = ring.length, j = n - 1; i < n; j = i++) {
    var pi = ring[i], xi = pi[0], yi = pi[1],
        pj = ring[j], xj = pj[0], yj = pj[1];
    if (((yi > y) ^ (yj > y)) && (x < (xj - xi) * (y - yi) / (yj - yi) + xi)) contains = !contains;
  }
  return contains;
}

function noop() {}

function shapefile_cancel() {
  return Promise.all([
    this._dbf && this._dbf.cancel(),
    this._shp.cancel()
  ]).then(noop);
}

function shapefile_read() {
  var that = this;
  return Promise.all([
    that._dbf && that._dbf.read(),
    that._shp.read()
  ]).then(function(results) {
    var dbf = results[0], shp = results[1];
    return shp.done ? shp : {
      done: false,
      value: {
        type: "Feature",
        properties: that._properties(dbf && dbf.value),
        geometry: that._geometry(shp.value)
      }
    };
  });
};

function shapefile(shpSource, dbfSource, decoder) {
  return Promise.all([
    shp(shpSource),
    dbfSource && dbf(dbfSource, decoder)
  ]).then(function(sources) {
    return new Shapefile(sources[0], sources[1]);
  });
}

function Shapefile(shp, dbf) {
  this._shp = shp;
  this._dbf = dbf;
  this._properties = dbf ? properties(dbf.fields) : noproperties;
  this._geometry = geometry(shp.shapeType);
  this.bbox = shp.box;
}

var prototype = Shapefile.prototype;
prototype.read = shapefile_read;
prototype.cancel = shapefile_cancel;

function open(shp, dbf, options) {
  if (typeof dbf === "string") {
    if (!/\.dbf$/.test(dbf)) dbf += ".dbf";
    dbf = path(dbf, options);
  } else if (dbf instanceof ArrayBuffer || dbf instanceof Uint8Array) {
    dbf = array(dbf);
  } else if (dbf != null) {
    dbf = stream(dbf);
  }
  if (typeof shp === "string") {
    if (!/\.shp$/.test(shp)) shp += ".shp";
    if (dbf === undefined) dbf = path(shp.substring(0, shp.length - 4) + ".dbf", options).catch(function() {});
    shp = path(shp, options);
  } else if (shp instanceof ArrayBuffer || shp instanceof Uint8Array) {
    shp = array(shp);
  } else {
    shp = stream(shp);
  }
  return Promise.all([shp, dbf]).then(function(sources) {
    var shp = sources[0], dbf = sources[1], encoding = "windows-1252";
    if (options && options.encoding != null) encoding = options.encoding;
    return shapefile(shp, dbf, dbf && new TextDecoder(encoding));
  });
}

function read(shp, dbf, options) {
  return open(shp, dbf, options).then(function(source) {
    var features = [], collection = {type: "FeatureCollection", features: features, bbox: source.bbox};
    return source.read().then(function read(result) {
      if (result.done) return collection;
      features.push(result.value);
      return source.read().then(read);
    });
  });
}

exports.open = open;
exports.read = read;